#ifndef __JABYENGINE_PRIMITIVE_RECTANGLE_TYPES_HPP__
#define __JABYENGINE_PRIMITIVE_RECTANGLE_TYPES_HPP__
#include "linked_elements.hpp"
#include "primitive_support_types.hpp"

namespace JabyEngine {
    namespace GPU {
        struct PagePositionClut {
            PagePosition page;
            PageClut     clut;
        };

        namespace internal {
            struct RectCode : public CodeBase<RectCode> {
                enum struct SizeType : uint8_t {
                    Variable    = 0,
                    Pixel1x1    = 1,
                    Sprite8x8   = 2,
                    Sprite16x16 = 3
                };
                static constexpr uint8_t CmdValue = 0b011;

                static constexpr auto Size = BitRange::from_to(27 - BitCorrection, 28 - BitCorrection);

                static constexpr RectCode create() {
                    return RectCode(CmdValue);
                }
            };

            template<typename T>
            struct RectCodeInterface {
                typedef typename ::JabyEngine::GPU::internal::RectCode Code;

                constexpr T& set_semi_transparent(bool set = true) {
                    static_cast<T*>(this)->code.set_tenary(set, Code::SemiTransparent, Code::NonTransparent);
                    return *static_cast<T*>(this);
                }
            };

            template<typename RectCode::SizeType Size> 
            struct RECT_BASE_F : public RectCodeInterface<RECT_BASE_F<Size>> {
                typedef RectCodeInterface<RECT_BASE_F<Size>>::Code Code;
                static constexpr auto IdentityCode = Code::create().set(Code::Size.with(static_cast<uint8_t>(Size))).set(Code::Untextured).set(Code::NonTransparent);

                Color24 color;
                Code    code;
                Vertex  position;   

                constexpr RECT_BASE_F() = default;
                constexpr RECT_BASE_F(const Vertex& position, const Color24& color) : color(color), code(IdentityCode), position(position) {
                }
            };

            template<typename RectCode::SizeType Size> 
            struct RECT_BASE_T : public RectCodeInterface<RECT_BASE_T<Size>> {
                typedef RectCodeInterface<RECT_BASE_T<Size>>::Code Code;
                static constexpr auto IdentityCode = Code(RECT_BASE_F<Size>::IdentityCode).set(Code::Textured);

                Color24      color;
                Code         code;
                Vertex       position;
                PagePosition page;
                PageClut     clut;

                constexpr RECT_BASE_T() = default;
                constexpr RECT_BASE_T(const Vertex& position, const PagePositionClut& page, const Color24& color = Color24::Grey()) : color(color), code(IdentityCode), position(position), page(page.page), clut(page.clut) {
                }
            };

            template<typename RectCode::SizeType Size> 
            struct RECT_F : public RECT_BASE_F<Size>, public internal::RenderPrimitive<RECT_F<Size>>, public internal::LinkedElementCreator<RECT_F<Size>> {
                using RECT_BASE_F<Size>::RECT_BASE_F;                
            };

            template<typename RectCode::SizeType Size> 
            struct RECT_T : public RECT_BASE_T<Size>, public internal::RenderPrimitive<RECT_T<Size>>, public internal::LinkedElementCreator<RECT_T<Size>> {
                using RECT_BASE_T<Size>::RECT_BASE_T;
            };
        }

        typedef internal::RECT_F<internal::RectCode::SizeType::Pixel1x1>    TILE_1;
        typedef internal::RECT_F<internal::RectCode::SizeType::Sprite8x8>   TILE_8;
        typedef internal::RECT_F<internal::RectCode::SizeType::Sprite16x16> TILE_16;
        struct TILE : public internal::RECT_BASE_F<internal::RectCode::SizeType::Variable>, public internal::RenderPrimitive<TILE>, public internal::LinkedElementCreator<TILE> {
            SizeI16 size;

            constexpr TILE() = default;
            constexpr TILE(const AreaI16& area, const Color24& color) : RECT_BASE_F(area.position, color), size(area.size) {
            }
        };

        typedef internal::RECT_T<internal::RectCode::SizeType::Pixel1x1>    SPRT_1;
        typedef internal::RECT_T<internal::RectCode::SizeType::Sprite8x8>   SPRT_8;
        typedef internal::RECT_T<internal::RectCode::SizeType::Sprite16x16> SPRT_16;
        struct SPRT : public internal::RECT_BASE_T<internal::RectCode::SizeType::Variable>, public internal::RenderPrimitive<SPRT>, public internal::LinkedElementCreator<SPRT> {
            SizeI16 size;

            constexpr SPRT() = default;
            constexpr SPRT(const AreaI16& area, const PagePositionClut& page, const Color24& color = Color24::Grey()) : RECT_BASE_T(area.position, page, color), size(area.size) {
            }
        };
    }
}

#endif //!__JABYENGINE_PRIMITIVE_RECTANGLE_TYPES_HPP__